/****************************************************************
* location      - pubdata or prvdata.
* table         - table name you are trying to display.
* pwd           - table password needed.
* secsub        - subset beans where clause (where statement) for the security subset.
* subclaus      - subset beans where clause (where statement).
* outname       - output data sets name (view3, etc.).
* outlib        - output libname (usually work).
* outtype       - table or view.
* keyvar        - the key variable from the table you want to display, used for merging.
* secmsgst      - message to display when sec. subset can not be applied.
* dnemsgst      - message to display when the table to display does not exist.
* bigcolst      - the big string that goes in the final sql step that defines all the columns.
******************************************************************/
%macro generate(location, table, pwd, secsub, subclaus, outname, outlib, outtype, keyvar, secmsgst, dnemsgst, bigcolst, baseds=empmast, obs=);
	%global debug;
        %local where dsname pass secwhere rc;
        %if &baseds = %then %let baseds=empmast;
        %if &debug ne Y %then %str(options nomprint nosymbolgen nomlogic;);
        %if %quote(&location)=%str() %then %let location=work;
        %let dsname=&location..&table;
        %let pass=pw="&pwd";
        %let secwhere=;
        %let where=;
        %if %length(%trim(&secsub)) > 0 %then %do;
                %let secwhere=where=(&secsub);
        %end;

        %if &obs ne %then
          %let obs=obs=&obs;

        %if &debug=Y %then %put NOTE: checking existences of &dsname;
        %if %sysfunc(exist(&dsname)) OR %sysfunc(exist(&dsname, view)) %then %do;
                * data set exists, so proceed.;
                * apply the security subset.;
                %let rc = 1;
                data _null_;
                        call symput("rc", "0");
                        stop;
                        set &dsname (&pass %unquote(&secwhere));
                run;
                options obs=max nosyntaxcheck replace;
                %if &rc = 1 %then %do;
			* subset was not successfully applied, try against master.;
			%if %length(%trim(&keyVar)) > 0 %then %do;
				%local tbfmtname;
				%let tbfmtname=%trim(&keyvar)_fmtname;
				
				%global &tbfmtname;
				* have a valid master column if it is needed.;
				%if %quote(%unquote(&&&tbfmtname)) = %str() %then %do;

					%let fmtname=%sysfunc(ranuni(0));                                                                                              
					%let fmtname=%sysevalf(&fmtname * 100000);                                                                              
					%let fmtname=f%sysfunc(int(&fmtname))f;  
				        %if &debug=Y %then %put NOTE: creating format: &fmtname using key variable &keyvar;
					
					data %unquote(temp&outname)(keep=&keyvar fmtname type label);
						if _N_ = 1 then call symput("rc", "2");
						set PUBDATA.&baseds (obs=max &pass %unquote(&secwhere));
						retain fmtname "&fmtname" type 'c' label 'keep';
						call symput("rc", "0");
					run;
					options obs=max nosyntaxcheck replace;

					%if &rc = 2 %then %do;
						proc format library=work;
							value $&fmtname ''='notkeep';
						run;
						%let rc = 0;
					%end;
					%if &rc = 0 %then %do;
						%let secsub=put(&keyVar,$&fmtname..)='keep';
						%let secwhere=where=(&secsub);
						proc freq data=%unquote(temp&outname)(rename=&keyVar=start);
							tables start * fmtname * type * label/noprint out=%unquote(temp&outname)(keep=start fmtname type label);
						run;

						proc format cntlin=%unquote(temp&outname) library=work;
						run;
					%end;
					
					%let &tbfmtname=&fmtname;
				%end;
				%else %do;
					* we can reuse the format.;
					%let fmtname=%unquote(&&&tbfmtname);
					%let secsub=put(&keyVar,$&fmtname..)='keep';
					%let secwhere=where=(&secsub);
					%if &debug=Y %then %put NOTE: reusing format &fmtname;
				%end;
				
				%let rc=1;
				data _null_;
					call symput("rc", "0");
					stop;
					set &dsname(&pass &secwhere);
				run;
			%end;
			%else
			   %if &debug=Y %then %put NOTE: could not apply security subset because there is no key variable;
                %end;
                %if &rc = 1 %then %do;
                        * could not apply security subset.;
        		%if &debug=Y %then %put NOTE: could not apply security subset: &secwhere;
                        data %unquote(temp&outname);
                                _HCMMSG = "&secmsgst";
                        run;
                        proc sql;
                                create &outtype %unquote(&outlib..&outname) as select *
                                from %unquote(temp&outname);
                        quit;
                %end;
                %else %do;
                
                        %if %length(%trim(&subclaus)) > 0 %then %do;
                                %let where=where (&subclaus);
                        %end;
                        %if %length(%trim(&secsub)) > 0 %then %do;
                                %if %length(%trim(&where)) > 0 %then %do;
                                        %let where=&where and (&secsub);
                                %end;
                                %else %do;
                                        %let where=where &secsub;
                                %end;
                        %end;
                        * got a valid data set created.;
                        %if &bigcolst = %str() %then %do;
				  data %unquote(temp&outname);
					  _HCMMSG = "&dnemsgst";
					run;
				  proc sql;
					  create &outtype %unquote(&outlib..&outname) as select *
					  from %unquote(temp&outname);
				  quit;
			  %end;
			  %else %do;
				  %if &debug=Y %then %put NOTE: building result with security subset: &secwhere;
				  %let bigcolst=%unquote(&bigcolst);
				  proc sql;
					  create &outtype %unquote(&outlib..&outname) as select
					  %unquote(%str(&bigcolst)
					  from &dsname (&obs &pass) %unquote(&where));
				  quit;
			  %end;
          	%end;
        %end;
        %else %do;
           %if &debug=Y %then %put ERROR: data set does not exist;
                * data set did not exist, put out fake data set/view with appropriate message in it.;
                data %unquote(%unquote(temp&outname));
                        _HCMMSG = "&dnemsgst";
                        output;
                run;

                proc sql;
                        create &outtype  &outlib..&outname as select *
                        from %unquote(temp&outname);
                quit;
        %end;
   options &mprint;
%mend;
