%macro secformat(baseds=pubdata.empmast,idvar=id,keepword=KEEP,fmtname=SEC1FMT,
                 lib=dwdata,etype=formatc,root=grplev,rootmax=,rootchck=);
                 
  %global hiermax debug;
  %if &debug ne Y %then %str(options nosource;);
	
  %local dsid secdate create fullname cat basedate success fmtdesc pos fmtdesc basevar basevalue tempname;

  /* create the full path for the format */
  %let cat=&lib..formats;
  %let fullname=&cat..&fmtname..&etype;
  %let tempname=work.formats.&fmtname..&etype;

  /* get the date of the security format catalog and place it into the secdate variable */
  %deltable(tables=catdate); 
  %let secdate=;
  data catdate;
    moddate=.;
    desc="";
  run;
  %if %sysfunc(cexist(&cat))=1 %then %do;
    proc catalog catalog=&cat entrytype=&etype;
      contents out=catdate(where=(name=upcase("&fmtname")));
    run;
  %end;
  data _null_;
    retain moddate .;
    set catdate;
    call symput('secdate',put(moddate,12.));
    call symput('fmtdesc',trim(desc));
    stop;
  run;

 /* %deltable(tables=catdate); */

  /* get the base variable and base value of the existing format */
  %if &fmtdesc ne %str() %then %do;
    %let pos=%index(&fmtdesc,=);
    %if &pos gt 0 %then %do;
      %let basevar=%trim(%scan(&fmtdesc,1,=));
      %let basevalue=%trim(%scan(&fmtdesc,2,=));
    %end;
  %end;

  /* get the date of the base data set */ 
  %let basedate=;
  %let dsid=%sysfunc(open(&baseds(&datapass)));
  %if &dsid gt 0 %then %do;
    %let basedate=%sysfunc(attrn(&dsid,MODTE));
    %let dsid=%sysfunc(close(&dsid));
  %end;

  /* if rootmax is blank, use hiermax */
  %if &rootmax = %then %let rootmax=&hiermax;

  /* set the create flag, -1=can't create 1=use existing 0=try to create */
  /* if there's any problem with the baseds or format, delete any existing format */
  %let create=1;
  
  /* the macro must have a root variable to base the format on */
  %if &rootchck = %then %do;
    %put WARNING: No base value to create security format.;
    %let create=-1;
  %end;

  %else %if %sysfunc(cexist(&fullname)) = 1 and 
       (&basevar ne &root or &basevalue ne &rootchck or &basevar= or &basevalue=) %then %do;
    %put WARNING: Base variable and value of existing format are not valid.;
    %let create=0;
 %end;

  /* Can't build if the dates on the base data set cannot be determined. */
  %else %if &basedate = %then %do;
    %put WARNING: Cannot determine if &baseds is valid, will recreate.;
    %let create=0;
  %end;
  
  /* if the format is newer than the base data set, then recreate */
  %else %if &basedate gt &secdate and %sysfunc(cexist(&fullname)) = 1 %then %do;
    %put NOTE: Existing security format is out of date.;
    %let create=0;
  %end;

  /* there must be a rootmax, if necessary and possible, use hiermax */
  %else %if &rootmax = %then %do;
    %put WARNING: There must be maximum value for the root of &root..;
    %let create=-1;
  %end;

  %if (&create=0 or &create=-1 or &basedate= or &secdate= or (&basedate gt &secdate)) and 
        %sysfunc(cexist(&fullname)) = 1  %then %do;
    %put NOTE: Deleting existing security format.;
    proc catalog cat=&cat;
      delete &fmtname / et=&etype;
    quit;
  %end;

  /* if the format is older than the base data set, then use existing */
  %else %if &basedate le &secdate and %sysfunc(cexist(&fullname)) = 1 %then %do;
    %put NOTE: Existing security format will be used.;
    %let create=1;
  %end;
  %else %let create=0;

  /* if possible, create a new security format */
  %if &create = 0 %then %do;

    /* check to see if you want to base the security format on the CWRKGRP table or the EMPMAST table */
    %global secformat_base;
    %if &secformat_base ne %str() %then %do;
  
      /* construct one macro variable for each group this person manages */
      %put NOTE: Building security format based on &secformat_base for ID=&id.;
      %do i=1 %to 100;%local rootchck&i;%end;
      data _null_;
        set &secformat_base(&datapass) end=end;
        if mgrempid="&id" then do;
          n+1;
          call symput('rootchck'!!left(put(n,4.)),trim(left(rptgrp)));
          %if &debug=Y %then %str(put 'NOTE: Found management group ' rptgrp= $20.;);
        end;
        if end then call symput('rootchckmax',left(put(n,4.)));
      run;
    %end;
    %else %do;
      %let rootchckmax=1;
      %let rootchck1=&rootchck;
    %end;
  
    /* now build the table that will have one record for each employee reporting to this person */
    %deltable(tables=reports)
    data reports;
      retain fmtname "$&fmtname" type 'C' label "&keepword";
      %do i=1 %to &rootchckmax;
        retain rootchck&i "&&rootchck&i";
      %end;
      set &baseds(&datapass keep=&idvar &root.1 - &root&rootmax) end=end;
      array rootvars(&rootmax) &root.1 - &root&rootmax;
      array rootchckvals(&rootchckmax) rootchck1-rootchck&rootchckmax;
      do i=1 to &rootmax until(found);
        do j=1 to &rootchckmax until(found);
          if rootvars(i)=rootchckvals(j) then do;
            ndata+1;
            output; 
            found=1;
            leave;
          end;
        end;
      end;
      if end and ndata gt 0 then call symput('SUCCESS','1');
      keep &idvar fmtname type label;
      rename &idvar=start;
    run; 
    
    /* if the reports dataset successfully built, then create a format from it */
    %if &success=1 %then %do;
      proc format library=&lib cntlin=reports;
      run; 
      proc catalog catalog=&cat;
	   modify &fmtname (description="&root=&rootchck") / entrytype=&etype;
      run;
      quit;
      %put NOTE: Security format has been created based on &root(x) equal to &rootchck.;
    %end;
    
    /* otherwise create a work format that will not allow any data to be shown */
    %else %do;
      %put WARNING: Unable to build security format. Using default format with no table access.;
      data reports;
        retain fmtname "$&fmtname" type 'C' label "NOKEEP" start '_nowayvalue_';
      run;
      proc format library=work cntlin=reports;
      run;		
    %end;
    %deltable(tables=reports);
  %end;

  /* if a permanent format did get created then put a description on it */
  %if %sysfunc(cexist(&fullname)) = 1 %then %do;
    options nonotes;
    proc catalog catalog=&cat;
      modify &fmtname (description="&root=&rootchck") / entrytype=&etype;
    run;
    quit;
    options notes;
  %end;
  
  /* otherwise if no format was created, not even the temporary work format, issue an error msg. */
  %else %if %sysfunc(cexist(&tempname)) ne 1 %then %do;
    %put ERROR: No security format.;
  %end;
%mend; 