<%-- Copyright (c) 2001 by SAS Institute Inc., Cary, NC 27513 --%>
<%@ page import="java.io.*,java.util.*" %>
<%@ page import="java.text.ParseException" %>
<%@ page import="com.sas.query.*" %>
<%@ page import="com.sas.query.app.*" %>
<%@ page import="com.sas.query.datasources.*" %>
<%@ page import="com.sas.query.models.*" %>
<%@ page import="com.sas.query.models.Expression" %>
<%@ page import="com.sas.query.parser.*" %>
<%@ page import="com.sas.query.visuals.*" %>
<%@ page import="com.sas.hrv.*" %>
<%@ page import="com.sas.rmi.Connection" %>
<%@ page import="com.sas.servlet.beans.*" %>
<%@ page import="com.sas.servlet.beans.html.*" %>
<%@ page import="com.sas.collection.StringCollection" %>
<%!

class OperandParser
{

// ### ADDITIONS FOR COLUMNNAME AS VALUE
	public String COLUMN_PREFIX = HRVParser.COLUMN_PREFIX;
// ### DONE WITH ADDITIONS

   HRVQuery m_query;

   public OperandParser(HRVQuery query)
   {
      this.m_query = query;
   }

   /**
    * Determines if the test string represents the value MISSING based both on
    * the content of the string and the data type of the expression in the
    * result item.
    */
   public boolean isMissingValue(ResultItem item, String test)
   {
      if (item == null)
      	 return isMissingValue((Expression)null, test);
      else
	 return isMissingValue(item.getExpression(), test);
   }

   /**
    * Determines if the test string represents the value MISSING based both on
    * the content of the string and the data type of the expression.
    */
   public boolean isMissingValue(Expression exp, String test)
   {
      //check the trivial case first
      if (test == null || test.trim().length() == 0)
      	 return true;

      //if the test string is not trivial, then check by data type
      if (exp != null) {
	 test = test.trim();
      	 switch (exp.getDataType()) {
	    case Expression.EXP_TYPE_CHARACTER:
	       if ("''".equals(test))			//CASE: ''
	       	  return true;
	       if ("\"\"".equals(test))			//CASE: ""
	       	  return true;
	       break;
	    case Expression.EXP_TYPE_DATE:
	       if ("''d".equalsIgnoreCase(test))	//CASE: ''d
		  return true;
	       break;
	    case Expression.EXP_TYPE_TIME:
	       if ("''t".equalsIgnoreCase(test))	//CASE: ''t
		  return true;
	       break;
	    case Expression.EXP_TYPE_TIMESTAMP:
	       if ("''dt".equalsIgnoreCase(test))	//CASE: ''dt
		  return true;
	       break;
	    case Expression.EXP_TYPE_NUMERIC:
	       if (".".equals(test))			//CASE: .
		  return true;
	       break;
	 }
      }

      return false;					//Must not be MISSING
   }

   protected HRVDataSourceParameter _getParameter(String name)
   {
      DataSourceParameterList parameters = ((HRVDataSource)m_query.getDataSource()).getParameters();
      for (int i=0; i<parameters.size(); i++) {
      	 DataSourceParameter p = parameters.elementAt(i);
	 if (p.getName().equalsIgnoreCase(name)) {
	    return (HRVDataSourceParameter)p;
	 }
      }
      return null;
   }

   public ExpressionList processOperand(int dataType, boolean isCommaList, String value)
      throws java.text.ParseException
   {
      ExpressionList expList = null;

      switch (dataType) {
      	 case Expression.EXP_TYPE_DATE:
	    expList = _processDateOperand(isCommaList, value);
	    break;
      	 case Expression.EXP_TYPE_TIME:
	    expList = _processTimeOperand(isCommaList, value);
	    break;
			 case Expression.EXP_TYPE_CHARACTER:
	    expList = _processCharOperand(isCommaList, value);
	    break;
      	 case Expression.EXP_TYPE_TIMESTAMP:
	    expList = _processTimestampOperand(isCommaList, value);
	    break;
      	 default:
	    break;
      }

      return expList;
   }

// ### ADDITIONS FOR COLUMNNAME AS VALUE
   public Expression _processColumnName(String value)
      throws java.text.ParseException
      //new ParseException(String message, int indexOfError)
   {
//System.out.println( "*** Entering _processColumnName( "+value+" )" );
//System.out.println( "   value.substring(1)="+value.substring(1) );
      Expression exp = null;
		if( value.startsWith(COLUMN_PREFIX))
		{
			ColumnList collist = m_query.getReferableColumns();
			boolean found = false;
			Column col = null;
			String testName = value.substring(1);
			for( int i = 0; !found && i < collist.size(); ++i )
			{
				col = collist.elementAt(i);
				String name = col.getName();
//System.out.println( "   name="+name );				
				if( testName.equalsIgnoreCase(name) )
					found = true;
				else
				{
			      SqlLogic logic = m_query.getSqlLogic();
					String fullname = logic.genColumnName( col );
//System.out.println( "   fullname="+fullname );					
					if( testName.equalsIgnoreCase(fullname) )
						found = true;
				}
			}

			if( found )
{
//System.out.println( "   Adding ColumnExpression" );
				exp = new ColumnExpression(col);
}
		}

//System.out.println( "--- Leaving _processColumnName ("+exp+")" );
		return exp;
	}
// ### DONE WITH ADDITIONS
	

   public ExpressionList _processDateOperand(boolean isCommaList, String value)
      throws java.text.ParseException
      //new ParseException(String message, int indexOfError)
   {
      ExpressionList expList = new ExpressionList();

      if (isCommaList) 
      {
			StringTokenizer st = new StringTokenizer(value, ",");
			for (;st.hasMoreTokens(); ) 
			{
				String token = st.nextToken();
// ### MODIFICATIONS FOR COLUMNNAME AS VALUE
		      Expression exp = _processColumnName(token);
		      if( exp != null )
					expList.addElement(exp);
		      else
					expList.addElement(_processSingleDateOperand(token));
// ### DONE WITH MODIFICATIONS
			}
		} 
		else 
		{
// ### MODIFICATIONS FOR COLUMNNAME AS VALUE
			Expression exp = _processColumnName(value);
			if( exp != null )
				expList.addElement(exp);
			else
				expList.addElement(_processSingleDateOperand(value));
// ### DONE WITH MODIFICATIONS
		}

      return expList;
   }

   public Expression _processSingleDateOperand(String value)
      throws java.text.ParseException
      //new ParseException(String message, int indexOfError)
   {
      Expression exp = null;

      //this is a single instance date operand. the value should be like:
      //    &[_a-zA-Z][_a-zA-Z0-9]*
      //    DATE7.
      //    DATE9.
      //    'DATE7.'
      //    'DATE9.'
      //    'DATE7.'d
      // or 'DATE9.'d

      value = value.trim();

      StringBuffer valueBuf = new StringBuffer(value.length());
      int index = 0;
      char ch = 0;
      int quotes = 0;
      int parseMode = 0;
      /*
       * 0=looking for whitespace, &, digit, or quote
       * 1=quote found - looking for digit or 2nd quote
       * 2=digit found - looking for digit or letter
       * 3=letter found - looking for letter or digit
       * 4=digit found - looking for 2nd quote, or end-of-string
       * 5=2nd quote found - looking for 'd','D', or end-of-string
       * 6='d' or 'D' found - looking for end-of-string
       * 7=& found - looking for [A-Za-z_]
       * 8=[A-Za-z_] found - looking for [A-Za-z0-9_] or end-of-string
       */
      while (index < value.length()) {
	 ch = value.charAt(index);
	 switch (parseMode) {
	    case 0:
	       if (Character.isWhitespace(ch)) {
		  index++;
	       } else if (Character.isDigit(ch)) {
		  parseMode = 2;
	       } else if (ch == '\'') {
		  parseMode = 1;
		  quotes++;
		  index++;
	       } else if (ch == '&') {
		  parseMode = 7;
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 1:
	       if (Character.isDigit(ch)) {
		  parseMode = 2;
	       } else if (ch == '\'') {
		  parseMode = 5;
		  quotes++;
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 2:
	       if (Character.isDigit(ch)) {
		  valueBuf.append(ch);
		  index++;
	       } else if (Character.isLetter(ch)) {
		  parseMode = 3;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 3:
	       if (Character.isLetter(ch)) {
		  valueBuf.append(ch);
		  index++;
	       } else if (Character.isDigit(ch)) {
		  parseMode = 4;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 4:
	       if (Character.isDigit(ch)) {
		  valueBuf.append(ch);
		  index++;
	       } else if (ch == '\'' && quotes == 1) {
		  parseMode = 5;
		  quotes++;
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 5:
	       if (ch == 'd' || ch == 'D') {
		  parseMode = 6;
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 6:
	       if (Character.isWhitespace(ch)) {
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 7:
	       if (ch == '_' || Character.isLetter(ch)) {
		  valueBuf.append(ch);
		  parseMode=8;
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 8:
	       if (ch == '_' || Character.isLetterOrDigit(ch)) {
		  valueBuf.append(ch);
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	 }
      }
      if ((parseMode == 4 || parseMode == 5 || parseMode == 6) && quotes != 1) {
	 exp = new ConstantExpression(valueBuf.toString(), Expression.EXP_TYPE_DATE);
      } else if (parseMode == 8) {
      	 HRVDataSourceParameter param = _getParameter(valueBuf.toString());
	 if (param != null)
	    exp = new ParameterExpression(param);
	 else
	    throw new ParseException("Invalid user prompt identifier!", index);
      } else {
	 throw new ParseException("Unexpected end of string while processing value!", index);
      }

      return exp;
   }

   public ExpressionList _processTimeOperand(boolean isCommaList, String value)
      throws java.text.ParseException
      //new ParseException(String message, int indexOfError)
   {
      ExpressionList expList = new ExpressionList();

      if (isCommaList) 
      {
			StringTokenizer st = new StringTokenizer(value, ",");
			for (;st.hasMoreTokens(); ) 
			{
				String token = st.nextToken();
// ### MODIFICATIONS FOR COLUMNNAME AS VALUE
				Expression exp = _processColumnName(token);
				if( exp != null )
					expList.addElement(exp);
				else
					expList.addElement(_processSingleTimeOperand(token));
// ### DONE WITH MODIFICATIONS
			}
		} 
		else 
		{
// ### MODIFICATIONS FOR COLUMNNAME AS VALUE
			Expression exp = _processColumnName(value);
			if( exp != null )
				expList.addElement(exp);
			else
				expList.addElement(_processSingleTimeOperand(value));
// ### DONE WITH MODIFICATIONS
		}

      return expList;
   }

   public Expression _processSingleTimeOperand(String value)
      throws java.text.ParseException
      //new ParseException(String message, int indexOfError)
   {
      Expression exp = null;

      //this is a single instance date operand. the value should be like:
      //    &[_a-zA-Z][_a-zA-Z0-9]*
      //    TIME.
      //    'TIME.'
      // or 'TIME.'t

      value = value.trim();

      StringBuffer valueBuf = new StringBuffer(value.length());
      int index = 0;
      char ch = 0;
      int quotes = 0;
      int parseMode = 0;
      /*
       * 0=looking for whitespace, negative, digit, &, or quote
       * 1=quote found - looking for negative, digit or 2nd quote
       * 2=digit found - looking for digit or colon
       * 3=colon found - looking for digit
       * 4=digit found - looking for 2nd colon
       * 5=colon found - looking for digit
       * 6=digit found - looking for digit, 2nd quote, or end-of-string
       * 7=2nd quote found - looking for 't','T', or end-of-string
       * 8='t' or 'T' found - looking for end-of-string
       * 9=& found - looking for [A-Za-z_]
       * 10=[A-Za-z_] found - looking for [A-Za-z_0-9] or end-of-string
       * 11=negative found - looking for digit
       */
      while (index < value.length()) {
	 ch = value.charAt(index);
	 switch (parseMode) {
	    case 0:
	       if (Character.isWhitespace(ch)) {
		  index++;
	       } else if (Character.isDigit(ch)) {
		  parseMode = 2;
	       } else if (ch == '-') {
		  parseMode = 11;
	       } else if (ch == '\'') {
		  parseMode = 1;
		  quotes++;
		  index++;
	       } else if (ch == '&') {
		  parseMode = 9;
		  quotes++;
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 1:
	       if (Character.isDigit(ch)) {
		  parseMode = 2;
	       } else if (ch == '-') {
		  parseMode = 11;
	       } else if (ch == '\'') {
		  parseMode = 5;
		  quotes++;
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 2:
	       if (Character.isDigit(ch)) {
		  valueBuf.append(ch);
		  index++;
	       } else if (ch == ':') {
		  parseMode = 3;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 3:
	       if (ch == ':') {
		  valueBuf.append(ch);
		  index++;
	       } else if (Character.isDigit(ch)) {
		  parseMode = 4;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 4:
	       if (Character.isDigit(ch)) {
		  valueBuf.append(ch);
		  index++;
	       } else if (ch == ':') {
		  parseMode = 5;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 5:
	       if (ch == ':') {
		  valueBuf.append(ch);
		  index++;
	       } else if (Character.isDigit(ch)) {
		  parseMode = 6;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 6:
	       if (Character.isDigit(ch)) {
		  valueBuf.append(ch);
		  index++;
	       } else if (ch == '\'' && quotes == 1) {
		  parseMode = 7;
		  quotes++;
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 7:
	       if (ch == 't' || ch == 'T') {
		  parseMode = 8;
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 8:
	       if (Character.isWhitespace(ch)) {
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 9:
	       if (ch == '_' || Character.isLetter(ch)) {
		  valueBuf.append(ch);
		  parseMode=10;
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 10:
	       if (ch == '_' || Character.isLetterOrDigit(ch)) {
		  valueBuf.append(ch);
		  index++;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	    case 11:
	       if (ch == '-') {
		  valueBuf.append(ch);
	       	  parseMode = 2;
		  index++;
	       } else if (Character.isDigit(ch)) {
	       	  parseMode = 2;
	       } else {
		  throw new ParseException("Illegal character encountered in value!", index);
	       }
	       break;
	 }
      }
      if ((parseMode == 6 || parseMode == 7 || parseMode == 8) && quotes != 1){
	 exp = new ConstantExpression(valueBuf.toString(), Expression.EXP_TYPE_TIME);
      } else if (parseMode == 10) {
      	 HRVDataSourceParameter param = _getParameter(valueBuf.toString());
	 if (param != null)
	    exp = new ParameterExpression(param);
	 else
	    throw new ParseException("Invalid user prompt identifier!", index);
      } else {
	 throw new ParseException("Unexpected end of string while processing value!", index);
      }

      return exp;
   }

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
   public ExpressionList _processCharOperand(boolean isCommaList, String value)
      throws java.text.ParseException
      //new ParseException(String message, int indexOfError)
   {

      ExpressionList expList = new ExpressionList();

      value = value.trim();

      //character expressions:
      //  single units -
      //	&[A-Za-z_][A-Za-z_0-9]*
      //	anything else
      Vector atomicUnits = new Vector();
      	 StringBuffer atomBuf = new StringBuffer();
      	 int parseMode = 0;
	 //parse modes:
	 // 0 - Looking for quote or chunk
	 // 1 - quote found, looking for matching quote, comma, or end-of-string
	 // 2 - chunk found, looking for comma or end-of-string
	 // 3 - matching quote found, check for escaped quote,comma, or EOS
	 char lastQuote = 0;
	 int quotes = 0;
	 StringTokenizer st1 = new StringTokenizer(value, "\"',", true);
	 while (st1.hasMoreTokens()) {
	    String chunk = st1.nextToken();
//application.log("Mode="+parseMode+", Chunk: (" + chunk + ")");
	    switch (parseMode) {
	       case 0:
		  chunk = chunk.trim();
		  if ("'".equals(chunk) || "\"".equals(chunk)) {
		     lastQuote = chunk.charAt(0);
		     quotes++;
		     parseMode = 1;
		  } else if (chunk.length() > 0) {
		     //chunk found
		     parseMode = 2;
		     atomBuf.append(chunk);
		  }
	       	  break;
	       case 1:
		  //search for matching quote that isn't escaped.
		  if (String.valueOf(lastQuote).equals(chunk)) {
		     quotes++;
		     parseMode = 3;
		  } else {
		     atomBuf.append(chunk);
		  }
	       	  break;
	       case 2:
		  //unquoted chunk found - add chunks until comma
		  if (",".equals(chunk) && isCommaList) 
		  {
				//end of atomic unit
				parseMode = 0;
				atomicUnits.addElement(atomBuf.toString());
				//process the atom - this one was unquoted
// ### MODIFICATIONS FOR COLUMNNAME AS VALUE
				Expression exp = _processColumnName(atomBuf.toString());
				if( exp != null )
					expList.addElement(exp);
				else
				{
					Expression e = _processSingleCharOperand(atomBuf.toString(), false);
					expList.addElement(e);
				}
// ### DONE WITH MODIFICATIONS
				atomBuf = new StringBuffer();
		  } else {
		     atomBuf.append(chunk);
		  }
		  break;
	       case 3:
		  if (String.valueOf(lastQuote).equals(chunk)) {
		     //escaped quote found
		     atomBuf.append(lastQuote);
		     quotes = 0;
		     parseMode = 1;
		  } else if (",".equals(chunk) && isCommaList) {
		     //end of atomic unit
		     quotes = 0;
		     lastQuote = 0;
		     parseMode = 0;
		     atomicUnits.addElement(atomBuf.toString());
		     //process the atom - this one was quoted
		     Expression e = _processSingleCharOperand(atomBuf.toString(), true);
		     expList.addElement(e);
		     atomBuf = new StringBuffer();
		  } else {
		     throw new ParseException("Unbalanced quote in value!", -1);
		  }
	       	  break;
	       default:

	    }
	 }
	 //check for valid atomic unit in buffer
	 if (atomBuf.length() > 0) {
	    atomicUnits.addElement(atomBuf.toString());
	    Expression e = _processSingleCharOperand(atomBuf.toString(), (quotes > 0));
	    expList.addElement(e);
	 }

/*
      for (int i=0; i<atomicUnits.size(); i++) {
      	 String unit = atomicUnits.elementAt(i).toString();
//application.log("Atomic Unit: " + unit);
	 Expression e = _processSingleCharOperand(unit);
	 expList.addElement(e);
      }
*/

      return expList;
   }

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
   public Expression _processSingleCharOperand(String value, boolean quoted)
      throws java.text.ParseException
      //new ParseException(String message, int indexOfError)
   {
      Expression exp = null;

		if (quoted) 
		{
			exp = new ConstantExpression(value, Expression.EXP_TYPE_CHARACTER);
		} 
		else 
		{
// ### MODIFICATIONS FOR COLUMNNAME AS VALUE
			exp = _processColumnName(value);
			if( exp == null )
			{
// ### DONE WITH MODIFICATIONS
				String parameterName = null;
				if (value.length() > 1 && value.startsWith("&")) 
				{
					parameterName = value.substring(1);
				}

				if (parameterName != null && ((HRVDataSource)m_query.getDataSource()).isValidSASName(parameterName))
				{
					HRVDataSourceParameter param = _getParameter(parameterName);
					if (param != null)
						exp = new ParameterExpression(param);
					else
						throw new ParseException("Invalid user prompt identifier!", -1);
				} 
				else 
				{
					exp = new ConstantExpression(value, Expression.EXP_TYPE_CHARACTER);
				}
			}
      }

      return exp;
   }

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
   public ExpressionList _processTimestampOperand(boolean isCommaList, String value)
      throws java.text.ParseException
      //new ParseException(String message, int indexOfError)
   {
      ExpressionList expList = new ExpressionList();

		if (isCommaList) 
		{
			StringTokenizer st = new StringTokenizer(value, ",");
			for (;st.hasMoreTokens(); ) 
			{
				String token = st.nextToken();
// ### MODIFICATIONS FOR COLUMNNAME AS VALUE
				Expression exp = _processColumnName(token);
				if( exp != null )
					expList.addElement(exp);
				else
					expList.addElement(_processSingleTimestampOperand(token));
// ### DONE WITH MODIFICATIONS
			}
		}
		else 
		{
// ### MODIFICATIONS FOR COLUMNNAME AS VALUE
			Expression exp = _processColumnName(value);
			if( exp != null )
				expList.addElement(exp);
			else
				expList.addElement(_processSingleTimestampOperand(value));
// ### DONE WITH MODIFICATIONS
		}

      return expList;
   }

   public Expression _processSingleTimestampOperand(String value)
      throws java.text.ParseException
   {
      Expression exp = null;

      //a single datetime should look like:
      //    &[_a-zA-Z][_a-zA-Z0-9]*
      //    DATETIME.
      //    'DATETIME.'
      // or 'DATETIME.'dt

      String trimmed = value.trim();
      int index = 0;
      while (Character.isWhitespace(trimmed.charAt(index))) {
      	 index++;
      }
      if (index > 0)
      	 trimmed = trimmed.substring(index);

      String parameterName = null;
      if (trimmed.length() > 1 && trimmed.startsWith("&")) {
	 parameterName = trimmed.substring(1);
      }

      if (parameterName != null && ((HRVDataSource)m_query.getDataSource()).isValidSASName(parameterName))
      {
	 HRVDataSourceParameter param = _getParameter(parameterName);
	 if (param != null)
	    exp = new ParameterExpression(param);
	 else
	    throw new ParseException("Invalid user prompt identifier!", -1);
      } else if (trimmed.startsWith("'")) {
      	 //quoted - should look like:
	 //    'DATETIME.
	 //    'DATETIME.'
	 // or 'DATETIME.'dt
	 trimmed = trimmed.substring(1);
	 int lastQuote = trimmed.lastIndexOf('\'');
	 if (lastQuote == -1) {
	    throw new ParseException("Unbalanced quote in value!", -1);
	 }
	 String ending = trimmed.substring(lastQuote);
	 if ("'".equals(ending) || "'dt".equalsIgnoreCase(ending)) {
	    trimmed = trimmed.substring(0, lastQuote);
	    exp = new ConstantExpression(trimmed, Expression.EXP_TYPE_TIMESTAMP);
	 } else {
	    throw new ParseException("Invalid DATETIME value!", -1);
	 }
      } else {
	 exp = new ConstantExpression(trimmed, Expression.EXP_TYPE_TIMESTAMP);
      }

      return exp;
   }

}
%>
